//Following two librarires are required because of bug in wifi module
//https://forum.arduino.cc/t/interesting-esp32-issue-cant-use-analogread-in-esp_wifimanager-library/679348
#include "esp32-hal-adc.h" // needed for adc pin reset
#include "soc/sens_reg.h" // needed for adc pin reset

//Camera and face recognition stuff
#include "esp_camera.h"
#include "fd_forward.h"
#include "fr_forward.h"
#include "fr_flash.h" //storage
#include "fb_gfx.h" //face squares

//Display
#include "SSD1306Wire.h"
#include "images.h" //smiley image

//MQTT
#include <PubSubClient.h>

//Web server
#include "esp_http_server.h"
#include <ArduinoWebsockets.h>
#include "camera_index.h"  


//--------------------------------------Pins--------------------------------------
/*
 5V - POWER
 GND - GND
 12 - Not used
 13 - Display SCL
 15 - Display SDA
 14 - Photo sensor
 2  - Relay
 4  - Led (not used)

 3V - VCC (Display and Relay and LDR)
 16 - Not used
 0  - Pull low for firmware upload
 GND - 
 VCC - Not used
 3 - FTDI (Tx)
 1 - FTDI (Rx)
 GND - GND (Display and Relay and LDR) 
 */

#define I2C_SDA 15 //green
#define I2C_SCL 13 //blue
#define RELAYPIN 2
#define PHOTOSENPIN 14


//--------------------------------------Display stuff--------------------------------------
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels
uint64_t reg_a;
OLEDDISPLAY_GEOMETRY g = GEOMETRY_128_64;
HW_I2C _i2cBus = I2C_ONE;
int freq = 700000;
SSD1306Wire display(0x3c, I2C_SDA, I2C_SCL,g, _i2cBus, freq);


//--------------------------------------My global vars--------------------------------------
unsigned long currentMillis = 0;
unsigned long lastHartbeat = 0;
unsigned long faceDetectedMills = 0;
unsigned long enrollmentStartedMills = 0;
unsigned long endDispensingAtMillis = 0;
unsigned long displaySetAttMillis = 0;
short int lightValue; //0-100
short int lightThreshold =50; //when light is detected, you can increase this if you have extremely bright room
unsigned long lastLightReport = 0;
bool faceDetected; //face id 0,1...
String detectedFaceName ="Casper";
short int hartbeatFreqSec = 1;//1 sec
short int faceTimeoutInterval = 7000; //7 sec
short int enrollmentInterval = 10000; //10 sec
short int coolOffInterval =7000;
short int messageTimeoutInterval = 2000;
short int guestInterval = 1000;
short int primaryInterval = 3000;
short int secondaryInterval = 1500;
short int dbgLevel = 1;//0 No logs, 1 Errors, 2 Info, 3 Verbose
String givingMode="strict"; //strict, normal , giveAll
bool dispensingInProgres = false;
String primaryFaceName = "Sweetie";
String secondaryFaceName = "Grumpy";
bool isEnrolling = false;
String globEnrollingName;
String displayL1="-->";
String displayL2="*";
bool refreshDisplay =true;
bool offlineMode = false;
bool readyToDispense = false;
double  progressBar = 0;


//--------------------------------------Wifi, MQTT and web server stuff--------------------------------------
#define hostName "CandyDisp"
char ssid[] = "changeit";            //  your network SSID (name) 
char pass[] = "changeit";       // your network password

#define mqtt_server "changeit"
#define mqtt_user "changeit"
#define mqtt_password "changeit"

#define control_topic "changeit"
#define hartbeat_topic "changeit"

#define log_ip_topic "changeit"
#define log_error_topic "changeit"
#define debug_topic "changeit"

#define PART_BOUNDARY "123456789000000000000987654321"
static const char* _STREAM_CONTENT_TYPE = "multipart/x-mixed-replace;boundary=" PART_BOUNDARY;
static const char* _STREAM_BOUNDARY = "\r\n--" PART_BOUNDARY "\r\n";
static const char* _STREAM_PART = "Content-Type: image/jpeg\r\nContent-Length: %u\r\n\r\n";
using namespace websockets;
WiFiClient  mqWifiClient;
PubSubClient mqClient(mqWifiClient);
WebsocketsClient webSocketClient;
WebsocketsServer socket_server;


//--------------------------------------Camera config--------------------------------------
httpd_handle_t camera_httpd = NULL;

#define PWDN_GPIO_NUM     32
#define RESET_GPIO_NUM    -1
#define XCLK_GPIO_NUM      0
#define SIOD_GPIO_NUM     26
#define SIOC_GPIO_NUM     27
#define Y9_GPIO_NUM       35
#define Y8_GPIO_NUM       34
#define Y7_GPIO_NUM       39
#define Y6_GPIO_NUM       36
#define Y5_GPIO_NUM       21
#define Y4_GPIO_NUM       19
#define Y3_GPIO_NUM       18
#define Y2_GPIO_NUM        5
#define VSYNC_GPIO_NUM    25
#define HREF_GPIO_NUM     23
#define PCLK_GPIO_NUM     22
#define ENROLL_CONFIRM_TIMES 5
#define FACE_ID_SAVE_NUMBER 7
#define FACE_COLOR_RED    0x000000FF
#define FACE_COLOR_GREEN  0x0000FF00
#define FACE_COLOR_YELLOW (FACE_COLOR_RED | FACE_COLOR_GREEN)


bool initCamera() {
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;
  //init with high specs to pre-allocate larger buffers
  if (psramFound()) {
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }
  esp_err_t result = esp_camera_init(&config);
  if (result != ESP_OK) {
    return false;
  }
  return true;
}




//--------------------------------------Face recognition config--------------------------------------
mtmn_config_t my_mtmn_init_config()
{
  mtmn_config_t mtmn_config = {0};
  mtmn_config.type = FAST;
  mtmn_config.min_face = 80;
  mtmn_config.pyramid = 0.707;
  mtmn_config.pyramid_times = 4;
  //mtmn_config.p_threshold.score = 0.6;
  mtmn_config.p_threshold.score = 0.55; //to increase detection/decrease accuracy 
  mtmn_config.p_threshold.nms = 0.7;
  mtmn_config.p_threshold.candidate_number = 20;
  //mtmn_config.r_threshold.score = 0.7; 
  mtmn_config.r_threshold.score = 0.65; //to increase detection/decrease accuracy
  mtmn_config.r_threshold.nms = 0.7;
  mtmn_config.r_threshold.candidate_number = 10;
  //mtmn_config.o_threshold.score = 0.7; 
  mtmn_config.o_threshold.score = 0.65; //to increase detection/decrease accuracy
  mtmn_config.o_threshold.nms = 0.7;
  mtmn_config.o_threshold.candidate_number = 1;
  return mtmn_config;
}

mtmn_config_t mtmn_config = my_mtmn_init_config();
face_id_name_list id_name_list = {0};
dl_matrix3du_t *image_matrix =  NULL;
camera_fb_t * fb = NULL;
dl_matrix3d_t *face_id = NULL;
dl_matrix3du_t *aligned_face = dl_matrix3du_alloc(1, FACE_WIDTH, FACE_HEIGHT, 3);



//--------------------------------------Web server--------------------------------------
static esp_err_t index_handler(httpd_req_t *req) {
  httpd_resp_set_type(req, "text/html");
  httpd_resp_set_hdr(req, "Content-Encoding", "gzip");
  return httpd_resp_send(req, (const char *)index_ov2640_html_gz, index_ov2640_html_gz_len);
}

httpd_uri_t index_uri = {
  .uri       = "/",
  .method    = HTTP_GET,
  .handler   = index_handler,
  .user_ctx  = NULL
};

void app_httpserver_init ()
{
  httpd_config_t config = HTTPD_DEFAULT_CONFIG();
  if (httpd_start(&camera_httpd, &config) == ESP_OK)
    Serial.println("httpd_start");
  {
    httpd_register_uri_handler(camera_httpd, &index_uri);
  }
}

//Web server runs in separate task on Core1
TaskHandle_t Task1;
void loop1(void * parameter) {
  while (1) {
    webSocketClient = socket_server.accept();
    vTaskDelay(10);
  }
  vTaskDelete(NULL);
}



//--------------------------------------Some char/string transfomations--------------------------------------

char* string2char(String myString){
    if(myString.length()!=0){
        char *p = const_cast<char*>(myString.c_str());
        return p;
    }
}
String ip2String(IPAddress ipa)
{
  String response;
  response+=(String)ipa[0]; response+=".";
  response+=(String)ipa[1]; response+=".";
  response+=(String)ipa[2]; response+=".";
  response+=(String)ipa[3];
  return response;
}


//--------------------------------------Display functions--------------------------------------
void setDispValue (String msg1, String msg2){
  if(msg1!=""){
    displayL1=msg1;
  }
  if(msg2!=""){
    displayL2=msg2;
  }
 refreshDisplay =true;
}

void toDisplay() {
  display.clear();
  display.setColor(WHITE);
  
  if (displayL2=="Smile"){
    drawSmile(); 
  }
  else {
    display.setFont(ArialMT_Plain_10);
    display.drawString(0, 0, displayL1);  
  
    display.setFont(ArialMT_Plain_16);
    if (displayL2.length()<=17){
      display.drawString(0, 14, displayL2);
    }
    else {
      display.drawString(0, 14, displayL2.substring(0,17));
      display.drawString(0, 31, displayL2.substring(17));
    }
    display.display();
  }
  refreshDisplay =false;
  displaySetAttMillis =currentMillis;
  
}

void drawSmile() {
    //https://convertio.co/png-xbm/
    display.clear();
    display.drawXbm(39, 0, smile_width, smile_height, smile_bits);
    display.display();
}

void drawFaceRecognisedBar(){
    progressBar = double(currentMillis-faceDetectedMills)/faceTimeoutInterval;
    progressBar = round(progressBar*100);
    display.drawProgressBar(0, 50, 128, 15, progressBar);
    display.display();
}
void drawEnrollmentBar(){
    progressBar = double(currentMillis-enrollmentStartedMills)/enrollmentInterval;
    progressBar = round(progressBar*100);
    display.drawProgressBar(0, 50, 128, 15, progressBar);
    display.display();
}




//--------------------------------------Logging--------------------------------------
void mqttLog(String msg){
  mqClient.publish(debug_topic, msg.c_str(), true);
}

void webLog(String msg){
  if (webSocketClient.available()){
    webSocketClient.send(msg);
  }
}

void myLog(String msg, short int logLevel, bool newLine){
  if (dbgLevel>=logLevel){
    if (newLine){
      Serial.println(msg);
    }
    else {
      Serial.print(msg);
    }
      setDispValue(msg,"");
      if (!offlineMode){
        mqttLog(msg);
        webLog(msg);
      }
    
  } 
}


//--------------------------------------Face boxes functions --------------------------------------
//draws boxes around faces on video stream
static void draw_face_boxes(dl_matrix3du_t *image_matrix, box_array_t *boxes, String recognisedFaceName){
    int x, y, w, h, i;
    uint32_t color = FACE_COLOR_YELLOW;
    if(recognisedFaceName =="Stranger"){
        color = FACE_COLOR_RED;
    } else if(recognisedFaceName != "Stranger" && recognisedFaceName != "" ){
        color = FACE_COLOR_GREEN;
    }
    fb_data_t fb;
    fb.width = image_matrix->w;
    fb.height = image_matrix->h;
    fb.data = image_matrix->item;
    fb.bytes_per_pixel = 3;
    fb.format = FB_BGR888;
    for (i = 0; i < boxes->len; i++){
        // rectangle box
        x = (int)boxes->box[i].box_p[0];
        y = (int)boxes->box[i].box_p[1];
        w = (int)boxes->box[i].box_p[2] - x + 1;
        h = (int)boxes->box[i].box_p[3] - y + 1;
        fb_gfx_drawFastHLine(&fb, x, y, w, color);
        fb_gfx_drawFastHLine(&fb, x, y+h-1, w, color);
        fb_gfx_drawFastVLine(&fb, x, y, h, color);
        fb_gfx_drawFastVLine(&fb, x+w-1, y, h, color);
#if 0
        // landmark
        int x0, y0, j;
        for (j = 0; j < 10; j+=2) {
            x0 = (int)boxes->landmark[i].landmark_p[j];
            y0 = (int)boxes->landmark[i].landmark_p[j+1];
            fb_gfx_fillRect(&fb, x0, y0, 3, 3, color);
        }
#endif
    }
}
static void rgb_print(dl_matrix3du_t *image_matrix, uint32_t color, const char * str){
    fb_data_t fb;
    fb.width = image_matrix->w;
    fb.height = image_matrix->h;
    fb.data = image_matrix->item;
    fb.bytes_per_pixel = 3;
    fb.format = FB_BGR888;
    fb_gfx_print(&fb, (fb.width - (strlen(str) * 14)) / 2, 10, color, str);
}

static int rgb_printf(dl_matrix3du_t *image_matrix, uint32_t color, const char *format, ...){
    char loc_buf[64];
    char * temp = loc_buf;
    int len;
    va_list arg;
    va_list copy;
    va_start(arg, format);
    va_copy(copy, arg);
    len = vsnprintf(loc_buf, sizeof(loc_buf), format, arg);
    va_end(copy);
    if(len >= sizeof(loc_buf)){
        temp = (char*)malloc(len+1);
        if(temp == NULL) {
            return 0;
        }
    }
    vsnprintf(temp, len+1, format, arg);
    va_end(arg);
    rgb_print(image_matrix, color, temp);
    if(len > 64){
        free(temp);
    }
    return len;
}


//--------------------------------------Face handling functions --------------------------------------
//The key function in the whole sketch.
//It looks for a face and tries to recognise it if found. Also used to learn/enroll new faces
static void recogniseFace() {
   size_t _jpg_buf_len = 0;
    uint8_t * _jpg_buf = NULL;
    String recognisedFaceName = "";
    fb = esp_camera_fb_get();
    _jpg_buf = fb->buf;
    _jpg_buf_len = fb->len;
    if (!fb) {
      myLog("Camera capture failed",1,true);
    }
    image_matrix = dl_matrix3du_alloc(1, fb->width, fb->height, 3);
    uint32_t res = fmt2rgb888(fb->buf, fb->len, fb->format, image_matrix->item);
    if (!res) {
      myLog("to rgb888 failed",1,true);
      dl_matrix3du_free(image_matrix);
    }

    box_array_t *net_boxes = face_detect(image_matrix, &mtmn_config);
    
    if (net_boxes) {
      if (align_face(net_boxes, image_matrix, aligned_face) == ESP_OK) {
        face_id = get_face_id(aligned_face);
        if (isEnrolling){
          char enrollingName[FACE_ID_SAVE_NUMBER * ENROLL_NAME_LEN] = {0,};
          globEnrollingName.toCharArray(enrollingName, sizeof(enrollingName));
          int8_t left_sample_face = enroll_face_id_to_flash_with_name(&id_name_list, face_id, enrollingName);
          if(left_sample_face == (ENROLL_CONFIRM_TIMES - 1)){
              myLog("Enrolling "+String(globEnrollingName),2,true);
          }
          myLog("Enrolling "+String(globEnrollingName)+", sample: "+ String(ENROLL_CONFIRM_TIMES - left_sample_face),2,true);
          enrollmentStartedMills = millis();
          if (left_sample_face == 0){
              isEnrolling = false;
              enrollmentStartedMills = millis()+10000;
              readyToDispense = true;
              myLog("Enrolled "+String(globEnrollingName),2,true);
              globEnrollingName ="";
          }
        }
        else {
          face_id_node *f  = recognize_face_with_name(&id_name_list, face_id);
          if (f) {
            myLog(String(f->id_name)+" recognised",3,true);
            setDispValue ("","Hi "+String(f->id_name)+"!");
            rgb_printf(image_matrix, FACE_COLOR_GREEN, f->id_name);
            recognisedFaceName = String(f->id_name);
            faceDetected = true;
            faceDetectedMills = currentMillis;
            if (String(f->id_name)==primaryFaceName){
              setDispValue ("","Smile");
            }
          }
          else if (!faceDetected){  
            //If face has been already recognised we don't want this to owerwrite it to Stranger if it fails recognition in the next frame
            myLog("Face not recognised",3,true);
            setDispValue ("","Hi Stranger!");
            rgb_print(image_matrix, FACE_COLOR_RED, "Stranger");
            recognisedFaceName ="Stranger";
            faceDetected = true;
            faceDetectedMills = currentMillis;
          }
          if (recognisedFaceName!=""){
            detectedFaceName = recognisedFaceName;
          }
        }
        dl_matrix3d_free(face_id);
      } else {
        myLog("Face Not Aligned",3,true);
      }
      draw_face_boxes(image_matrix, net_boxes, recognisedFaceName);
      free(net_boxes->score);
      free(net_boxes->box);
      free(net_boxes->landmark);
      free(net_boxes);
  
      if (!fmt2jpg(image_matrix->item, fb->width * fb->height * 3, fb->width, fb->height, PIXFORMAT_RGB888, 90, &_jpg_buf, &_jpg_buf_len)){
        myLog("fmt2jpg failed",1,true);
      }
      esp_camera_fb_return(fb);
      fb = NULL;
    }
    dl_matrix3du_free(image_matrix);
  
    if (webSocketClient.available()){
      webSocketClient.sendBinary((const char *)_jpg_buf, _jpg_buf_len);
    }
    if (fb){
      esp_camera_fb_return(fb);
      fb = NULL;
      _jpg_buf = NULL;
    } 
    else if(_jpg_buf){
      free(_jpg_buf);
      _jpg_buf = NULL;
    }  
}

//face monitoring task
TaskHandle_t Task0;
void loop0(void * parameter) {
  while (1) {
    recogniseFace();
    vTaskDelay(10);
  }
  vTaskDelete(NULL);
}

//deletes face
int deleteFace(String faceName){
  char deletingName[FACE_ID_SAVE_NUMBER * ENROLL_NAME_LEN];
  faceName.toCharArray(deletingName, sizeof(deletingName));
  uint8_t tmpCount = id_name_list.count;
  delete_face_id_in_flash_with_name(&id_name_list, deletingName);
  if (tmpCount>id_name_list.count){
    myLog(String (faceName)+" deleted",1,true);
  }
  else{
    myLog("Deleting failed",1,true);
  } 
}

//--------------------------------------MQTT connection and message processing functions--------------------------------------
//To connect and reconnect in case of dropped connection
bool reconnectMessage() {
  // Loop until we're reconnected
  int connectAttempts = 0;
  while (!mqClient.connected()) {
    myLog("Attempting MQTT",2,true);
    if (mqClient.connect(hostName, mqtt_user, mqtt_password)) {
      delay(100);
      mqClient.subscribe(control_topic);
      myLog("Subscribed to MQTT ",2,true);
      return true;
    } else {      
      myLog("Message failed, rc="+String(mqClient.state()),1,true);
      if (connectAttempts>0) {
        return false;
      }      
      myLog(" try again in 5 seconds",1,false);
      // Wait 5 seconds before retrying
      delay(5000);
      connectAttempts++;

    }
  }
}


//MQTT commands
void process_command (String command, String value) {
  char* primCharArrayName = string2char(primaryFaceName);
  char* secCharArrayName = string2char(secondaryFaceName);
  
  if (command=="enrollPrimaryFace"){
    isEnrolling = true;
    enrollmentStartedMills = millis();
    readyToDispense = false;
    myLog("About to enroll primary",1,true);
    globEnrollingName = primaryFaceName;
  }
  else if (command=="enrollSecondaryFace"){
    isEnrolling = true;
    enrollmentStartedMills = millis();
    readyToDispense = false;
    myLog("About to enroll secondary",3,true);
    globEnrollingName = secondaryFaceName;
  }
  else if (command=="deletePrimaryFace"){
    myLog("Deleting primary face",1,true);
    deleteFace(primaryFaceName);
  }
  else if (command=="deleteSecondaryFace"){
    myLog("Deleting secondary face",3,true);
    deleteFace(secondaryFaceName);
  }
  else if (command=="deleteFaces"){
    myLog("Deleting all faces",1,true);
    deleteFace(primaryFaceName);
    deleteFace(secondaryFaceName);
  }
  else if (command=="strictMode"){
    myLog("Switching to strict mode",2,true);
    givingMode = "strict";
  }
  else if (command=="normalMode"){
    myLog("Switching to normal mode",2,true);
    givingMode = "normal";
  }
  else if (command=="giveAllMode"){
    myLog("Switching to give all mode",2,true);
    givingMode = "giveAll";
  }
  else if (command=="dbgLevel"){
    myLog("Changing debug level to "+value,2,true);
    dbgLevel = value.toInt();
  }
  else if (command=="guestInterval"){
    myLog("Changing guest interval to "+value,2,true);
    guestInterval = value.toInt();
  }
  else if (command=="primaryInterval"){
    myLog("Changing primary interval to "+value,2,true);
    primaryInterval = value.toInt();
  }
  else if (command=="secondaryInterval"){
    myLog("Changing secondary interval to "+value,3,true);
    secondaryInterval = value.toInt();
  }
  else if (command=="faceTimeoutInterval"){
    myLog("Changing face valid interval to "+value,2,true);
    faceTimeoutInterval = value.toInt();
  }
  else if (command=="lightThreshold"){
    myLog("Changing light threshold "+value,2,true);
    lightThreshold = value.toInt();
  }
  else if (command=="primaryFaceName"){
    myLog("Changing primary face name to "+value,2,true);
    primaryFaceName = value;
  }
  else if (command=="secondaryFaceName"){
    myLog("Changing secondary face name to "+value,3,true);
    secondaryFaceName = value;
  }
  else if (command=="Smile"){
    //Disenser will smile at you
    drawSmile();
    delay(10000);
  }  
  else myLog("Invalid command",2,true);

}

//MQTT message listener
void callbackMessage(char* topic, byte* message, unsigned int length) {
  if (String(topic) == control_topic) {
    myLog("Callback",3,true);
    String command;
    String value;
    String delim = "=";
    bool nowReadValue = false;
    for (int i = 0; i < length; i++) {
      if ((char)message[i]==(char)delim[0]){
        nowReadValue = true;
        continue;
      }
      if (!nowReadValue) {
        command += (char)message[i];
      }
      else {
        value += (char)message[i];
      }
    }
    myLog(command+": ",3,false);
    myLog(value,3,true);
    if (command.length()>0 && value.length()>0){
      process_command(command, value);
    }
  }
}

//MQTT hartbeat telling Home Assistant that dispenser is alive, i.e. it hasn't crashed or something
void hartbeat(){
  if (currentMillis - lastHartbeat > hartbeatFreqSec*1000){
    lastHartbeat = currentMillis;
    mqClient.publish(hartbeat_topic, "On", true);
  }
}




//--------------------------------------Dispensing functions--------------------------------------
void startDispensing(short int mills){
  dispensingInProgres = true;
  readyToDispense = false;
  digitalWrite(RELAYPIN, HIGH);
  endDispensingAtMillis = currentMillis + mills;
}

void stopDispensing(){
  digitalWrite(RELAYPIN, LOW);
  dispensingInProgres = false;
}


//Function to detect light, i.e. hand proximity
int checkLight (){
  //There is a bug in the wifi module and as a workaround we need to write back registers before analogRead
  //https://forum.arduino.cc/t/interesting-esp32-issue-cant-use-analogread-in-esp_wifimanager-library/679348
  WRITE_PERI_REG(SENS_SAR_READ_CTRL2_REG, reg_a);
  SET_PERI_REG_MASK(SENS_SAR_READ_CTRL2_REG, SENS_SAR2_DATA_INV);
  int lightVal=map(analogRead(PHOTOSENPIN), 0, 4095, 0, 100);
  
  if (lightVal==100){
    lightVal=0; // to cater for sensor errors
  }
  return  lightVal;
}

//logic that determines if candy will be given or not
void handleLightOn (){
  myLog("Handling "+detectedFaceName,3,true);
  if (faceDetected){
    if (detectedFaceName=="Stranger"){
      if(givingMode != "strict"){
        setDispValue ("","Take one stranger");
        startDispensing(guestInterval);
      }
      else {
        setDispValue ("","I don't like     strangers");
      }
    }
    else if (detectedFaceName==primaryFaceName){
      //showing smiley here
      startDispensing(primaryInterval);
    }
    else if (detectedFaceName==secondaryFaceName){
      setDispValue ("","You'll get fat!");
      startDispensing(secondaryInterval);
    }
    else {
      setDispValue ("","I don't know you");
    }
  }
  else if (givingMode == "giveAll"){
    setDispValue("","Take a candy      Casper");
    startDispensing(guestInterval);
  }
  else {
    setDispValue ("","Sorry Casper, no candy for you");
  }
}


//Function that runs just once when disopenser is turned on
void setup() {
  Serial.begin(115200);
  pinMode(RELAYPIN, OUTPUT);
  pinMode(PHOTOSENPIN, INPUT); 
  Serial.setDebugOutput(true);
  display.init();
  display.flipScreenVertically();
  display.clear();
  delay(2000);
  myLog("Booting",3, true);
  toDisplay();
  delay(1000);

  
  if (!initCamera()) {
 
    myLog("Failed to initialize camera...",1,true);
    toDisplay();
    delay(1000);
    return;
  }
  myLog("Camera initialized.",2,true);
  
  //drop down frame size for higher initial frame rate
  sensor_t * s = esp_camera_sensor_get();
  s->set_framesize(s, FRAMESIZE_QVGA);
  s->set_vflip(s, 1); //vertical flip
 

  //Need to store reg value because of wifi bug that prevents analogRead
  reg_a = READ_PERI_REG(SENS_SAR_READ_CTRL2_REG);
  WiFi.mode(WIFI_STA);
  //WiFi.setSleep(false); //disables wifi power saving and allegedly increase wifi speed. Haven't noticed any difference
  WiFi.begin(ssid, pass);
  short int retriesLeft = 5;
  while (WiFi.status() != WL_CONNECTED && retriesLeft>0) {
      delay(500);
      myLog(".",3, false);
      retriesLeft--;
  }
  if (WiFi.status() != WL_CONNECTED){
    offlineMode = true;
  }

  if (!offlineMode){
    myLog (ip2String(WiFi.localIP()).c_str(),1,true);
    toDisplay();
    delay(2000);
    app_httpserver_init();
  
    mqClient.setServer(mqtt_server, 1883);
    mqClient.setCallback(callbackMessage);
    delay(100);
  
      if (reconnectMessage()){
        socket_server.listen(82);
        if (socket_server.available()){
          myLog ("Server live:"+ip2String(WiFi.localIP()),2,true);
        }
    
        xTaskCreatePinnedToCore(
            loop1, 
            "Task1", 
            3000,  //uses ~2k
            NULL,
            0, 
            &Task1, 
            1);
      }
      else {
        offlineMode = true;
      }
  }

  xTaskCreatePinnedToCore(
      loop0, /* Function to implement the task */
      "Task0", /* Name of the task */
      4000, /* Stack size in words */ //usess ~3k
      NULL, /* Task input parameter */
      1, /* Priority of the task */
      &Task0, /* Task handle. */
      0); /* Core where the task should run */
  
  // Read current face data from on-board flash
  face_id_name_init(&id_name_list, FACE_ID_SAVE_NUMBER, ENROLL_CONFIRM_TIMES);
  short int noFaces =0;
  noFaces = read_face_id_from_flash_with_name(&id_name_list);
  if(noFaces>0){
    myLog(String(noFaces)+" faces fetched",3,true);
    readyToDispense = true;
  }
  else {
    myLog("No faces fetched",2,true);
    process_command ("enrollPrimaryFace", "");
  }
  toDisplay();
  delay(1000);

  if (offlineMode) {
    myLog("Offline mode",2, true);
    toDisplay();
    delay(1000);
  }
}


//handy for debuging memory leaks, disabled by default
void memMonitor(){
    Serial.print(currentMillis/1000/60);
    Serial.print("--------------------------------->");
    //Serial.print(xPortGetCoreID());
    Serial.print("Heap: ");
    Serial.print(ESP.getMaxAllocHeap()/1024);
    Serial.print(" - ");
    Serial.print(ESP.getFreeHeap()/1024);
    Serial.print("   S0: ");
    Serial.print(uxTaskGetStackHighWaterMark(Task0));
    Serial.print("   S1: ");
    Serial.print(uxTaskGetStackHighWaterMark(Task1));
    Serial.print(" PSRAM: ");
    Serial.print(ESP.getPsramSize()/1024);    
    Serial.print(" - ");
    Serial.print(ESP.getFreePsram()/1024); 
    Serial.println("");
  
}


//Main loop
//Keeps track on time and raises or lowers flags (dispensing, face validity, etc.)
//Refreshes display
//Handles MQTT conncetion

void loop() {
  //memMonitor(); //handy for debuging memory leaks, disabled by default

  if (refreshDisplay){
    toDisplay();
    refreshDisplay=false;
  }

  currentMillis = millis();
  
  if (isEnrolling){
    drawEnrollmentBar();
    if ((currentMillis-enrollmentStartedMills)>enrollmentInterval){
      isEnrolling=false;
      myLog("Enrollment ended",2,true);
      readyToDispense = true;
      
    }
  }
  if (readyToDispense){
    lightValue = checkLight();
    if (lightValue>lightThreshold && !dispensingInProgres){
      handleLightOn();
    }
    if (faceDetected){
      if(currentMillis<faceDetectedMills+faceTimeoutInterval){
        if (detectedFaceName!="Stranger" || givingMode != "strict"){
          drawFaceRecognisedBar();
        }
        
      }
      else{
        myLog("Face expired",3,true);
        faceDetected = false;
        detectedFaceName = "Casper";
      }
    }
    else if (currentMillis>displaySetAttMillis+messageTimeoutInterval) {
      setDispValue (" ","Candy anyone?");
    }
  }
  else if (currentMillis>endDispensingAtMillis+coolOffInterval && !isEnrolling){
    readyToDispense = true;
  }
  
  if (dispensingInProgres){ 
    if (currentMillis>=endDispensingAtMillis){
      stopDispensing();
      faceDetected = false;
      detectedFaceName = "Casper";
    }
  }

  if (!offlineMode){
    if (!mqClient.connected()) {
      reconnectMessage();
    }
    mqClient.loop();
    hartbeat();
  }

  delay(100);
}